local M = {}
M.type = "auxiliary"
M.relevantDevice = "engine"

local abs = math.abs
local min = math.min
local max = math.max
local rpmToAV = 0.104719755

local engine = nil
local p1_Motors = nil
local p3_Motors = nil
local motors = nil
local throttleSet = nil
local freeRPM = 6000
local rpm1 = 1500
local rpm2 = 2000
local rpm3 = 3000
local speed1 = 60
local speed2 = 100
local idlingTime = 0
local idlingBreak = 0
local lastBatteryLevel = 0

local function updateMotorIgnition()
  if electrics.values.ignitionLevel >= 2 then
    for _, motor in ipairs(p1_Motors) do
      motor.ignitionCoef = 1
    end
    for _, motor in ipairs(p3_Motors) do
      motor.ignitionCoef = 1
    end
    for _, motor in ipairs(motors) do
      motor.ignitionCoef = 1
    end
  elseif electrics.values.ignitionLevel < 2 then
    for _, motor in ipairs(p1_Motors) do
      motor.ignitionCoef = 0
    end
    for _, motor in ipairs(p3_Motors) do
      motor.ignitionCoef = 0
    end
    for _, motor in ipairs(motors) do
      motor.ignitionCoef = 0
    end
  end
end

local function updateEHSState(dt)
  electrics.values.checkRE = engine.isDisabled and 1 or 0
  local function startEngine()
    engine.electricsThrottleName = "engineTrottle"
    if electrics.values.reRunning == 0 and electrics.values.fuel < electrics.values.hybridSOC / 100 * 0.95 then
      if engine.outputAV1 < rpm1 * rpmToAV * 0.8 then
        engine:deactivateStarter(engine)
        electrics.values.p1RegenThrottle = 0
        electrics.values.p1Throttle = 0.5
      else
        engine:deactivateStarter(engine)
        engine.ignitionCoef = 1
        electrics.values.reRunning = 1
        electrics.values.engineTrottle = throttleSet
        electrics.values.p1Throttle = 0
      end
    end
  end

  local function stopEngine()
    engine:deactivateStarter(engine)
    electrics.values.engineTrottle = 0
    engine.ignitionCoef = 0
    electrics.values.reRunning = 0
    electrics.values.p1Throttle = 0
    electrics.values.p1RegenThrottle = 0
  end

  local function idlingTempKeeper()
    if engine.thermals.coolantTemperature > 90 then
      idlingBreak = 1
    end
    if idlingBreak == 1 and engine.thermals.coolantTemperature < 88 then
      idlingBreak = 0
    end
  end

  if electrics.values.checkRE == 1 then
    electrics.values.hybridMode = "EV"
  end
  if electrics.values.ignitionLevel == 2 and electrics.values.checkRE == 0 and electrics.values.fuel < 0.15 then
    electrics.values.hybridMode = "RE"
  end

  if electrics.values.ignitionLevel < 2 or electrics.values.hybridMode == "EV" or electrics.values.fuel >= electrics.values.hybridSOC / 100 or engine.thermals.coolantTemperature >= 110 then
    idlingTime = 0
    stopEngine()
  elseif electrics.values.ignitionLevel == 2 and electrics.values.hybridMode == "RE" then
    idlingTempKeeper()
    if electrics.values.wheelspeed * 3.6 <= 15 and electrics.values.fuel > electrics.values.hybridSOC / 100 * 0.5 and idlingBreak == 1 then
      engine:setTempRevLimiter(rpm1 * rpmToAV)
      idlingTime = idlingTime + dt
      if idlingTime >= 15 then
        stopEngine()
      else
        if electrics.values.reRunning == 1 then
          electrics.values.engineTrottle = throttleSet
          if engine.outputAV1 < engine.tempRevLimiterAV * 0.9 then
            electrics.values.p1RegenThrottle = max(0, electrics.values.p1RegenThrottle - 0.02)
          end
          if engine.throttle < throttleSet then
            electrics.values.p1RegenThrottle = min(1, electrics.values.p1RegenThrottle + 0.02)
          end
        else
          electrics.values.engineTrottle = 0
        end
      end
    elseif electrics.values.wheelspeed * 3.6 < speed1 or engine.thermals.coolantTemperature > 100 then
      idlingTime = 0
      engine:setTempRevLimiter(rpm1 * rpmToAV)
      if engine.thermals.coolantTemperature < 105 then
        startEngine()
      end
      if electrics.values.reRunning == 1 then
        electrics.values.engineTrottle = throttleSet
        if engine.outputAV1 < engine.tempRevLimiterAV * 0.9 then
          electrics.values.p1RegenThrottle = max(0, electrics.values.p1RegenThrottle - 0.02)
        end
        if engine.throttle < throttleSet then
          electrics.values.p1RegenThrottle = min(1, electrics.values.p1RegenThrottle + 0.02)
        end
      else
        electrics.values.engineTrottle = 0
      end
    elseif electrics.values.wheelspeed * 3.6 >= speed2 or electrics.values.fuel <= 0.1 then
      idlingTime = 0
      engine:setTempRevLimiter(rpm3 * rpmToAV)
      startEngine()
      if electrics.values.reRunning == 1 then
        electrics.values.engineTrottle = throttleSet
        if engine.outputAV1 < engine.tempRevLimiterAV * 0.9 then
          electrics.values.p1RegenThrottle = max(0, electrics.values.p1RegenThrottle - 0.02)
        end
        if engine.throttle < throttleSet then
          electrics.values.p1RegenThrottle = min(1, electrics.values.p1RegenThrottle + 0.02)
        end
      else
        electrics.values.engineTrottle = 0
      end
    elseif electrics.values.wheelspeed * 3.6 >= speed1 and electrics.values.wheelspeed * 3.6 < speed2 then
      idlingTime = 0
      engine:setTempRevLimiter((rpm2 + (rpm3 - rpm2) * (1 - (speed2 - electrics.values.wheelspeed * 3.6) / (speed2 - speed1))) * rpmToAV)
      startEngine()
      if electrics.values.reRunning == 1 then
        electrics.values.engineTrottle = throttleSet
        if engine.outputAV1 < engine.tempRevLimiterAV * 0.9 then
          electrics.values.p1RegenThrottle = max(0, electrics.values.p1RegenThrottle - 0.02)
        end
        if engine.throttle < throttleSet then
          electrics.values.p1RegenThrottle = min(1, electrics.values.p1RegenThrottle + 0.02)
        end
      else
        electrics.values.engineTrottle = 0
      end
    end
  end
end

local function updateGaugeInfo(dt)
  local totalCapacity = 0
  local remainingEnergy = 0
  
  for _, s in pairs(engine.registeredEnergyStorages) do
    local storage = energyStorage.getStorage(s)
    if storage then
      totalCapacity = totalCapacity + storage.energyCapacity
      remainingEnergy = remainingEnergy + storage.storedEnergy
    end
  end
  local remainingGasolineRatio = remainingEnergy / totalCapacity
  electrics.values.remainingGasolineRatio = remainingGasolineRatio
  if remainingGasolineRatio <= 0.1 then
    electrics.values.lowGasoline = 1
  else
    electrics.values.lowGasoline = 0
  end

  local motorThrottle = 0
  local motorRegenThrottle = 0
  local motorCount = #p3_Motors + #motors
  electrics.values.motorCount = motorCount
  for _, motor in ipairs(p3_Motors) do
    motorThrottle = motorThrottle + electrics.values[motor.electricsThrottleName]
    motorRegenThrottle = motorRegenThrottle + electrics.values[motor.electricsRegenThrottleName]
  end
  for _, motor in ipairs(motors) do
    motorThrottle = motorThrottle + electrics.values[motor.electricsThrottleName]
    motorRegenThrottle = motorRegenThrottle + electrics.values[motor.electricsRegenThrottleName]
  end
  motorThrottle = motorThrottle / motorCount
  motorRegenThrottle = motorRegenThrottle / motorCount
  
  if electrics.values.reRunning == 1 and electrics.values.wheelspeed >= 0.1 then
    if motorThrottle > 0 then
      if electrics.values.fuel < lastBatteryLevel then
        electrics.values.hybridEnergyFlow = "S-"
      elseif electrics.values.fuel == lastBatteryLevel then
        electrics.values.hybridEnergyFlow = "SS"
      elseif electrics.values.fuel > lastBatteryLevel then
        electrics.values.hybridEnergyFlow = "S+"
      end
    elseif motorThrottle <= 0 and motorRegenThrottle >= 0 and motorRegenThrottle < 1e-2 then
      electrics.values.hybridEnergyFlow = "SC"
    elseif motorThrottle <= 0 and motorRegenThrottle >= 1e-2 then
      electrics.values.hybridEnergyFlow = "SR"
    end
  elseif electrics.values.reRunning == 0 and electrics.values.wheelspeed >= 0.1 then
    if motorThrottle > 0 then
      electrics.values.hybridEnergyFlow = "E-"
    elseif motorThrottle <= 0 and motorRegenThrottle >= 0 and motorRegenThrottle < 1e-2 then
      electrics.values.hybridEnergyFlow = "EE"
    elseif motorThrottle <= 0 and motorRegenThrottle >= 1e-2 then
      electrics.values.hybridEnergyFlow = "ER"
    end
  elseif electrics.values.wheelspeed < 0.1 then
    if electrics.values.reRunning == 1 then
      electrics.values.hybridEnergyFlow = "SC"
    else
      electrics.values.hybridEnergyFlow = "EE"
    end
  end
  lastBatteryLevel = electrics.values.fuel
end

local function updateGFX(dt)
  updateMotorIgnition()
  updateEHSState(dt)
  updateGaugeInfo(dt)
end

local function serialize()
end

local function deserialize(data)
end

local function init(jbeamData)
  extensions.load("socControl")
  engine = powertrain.getDevice(jbeamData.engineName) or powertrain.getDevice("mainEngine")
  p1_Motors = powertrain.getDevicesByType("p1_Motor")
  motors = powertrain.getDevicesByType("electricMotor")
  p3_Motors = powertrain.getDevicesByType("p3_Motor")
  --throttleSet = jbeamData.throttleSet or 0.7
  freeRPM = engine.maxRPM or 6000
  rpm1 = jbeamData.rpm1 or 1500
  rpm2 = jbeamData.rpm2 or 2000
  rpm3 = jbeamData.rpm3 or 3000
  speed1 = jbeamData.speed1 or 60
  speed2 = jbeamData.speed2 or 100
  electrics.values.checkRE = engine.isDisabled and 1 or 0

  local minV = nil
  for k, v in pairs(engine.invBurnEfficiencyTable) do
    if minV == nil or v < minV or (v == minV and k / 100 < throttleSet) then
      minV = v
      throttleSet = k / 100
    end
  end
end

local function initLastStage(jbeamData)
end

M.init = init
M.initLastStage = initLastStage
M.updateMotorIgnition = updateMotorIgnition
M.updateEHSState = updateEHSState
M.updateGFX = updateGFX
M.serialize = serialize
M.deserialize = deserialize

return M
