local M = {}
M.type = "auxiliary"
M.relevantDevice = "engine"

local abs = math.abs
local min = math.min
local max = math.max
local rpmToAV = 0.104719755

local engine = nil
local iDM_Transmission = nil
local p1_Motors = nil
local p3_Motors = nil
local motors = nil
local throttleSet = nil
local freeRPM = 6000
local rpm1 = 1500
local rpm2 = 2000
local rpm3 = 3000
local speed1 = 60
local speed2 = 100
local idlingTime = 0
local idlingBreak = 0
local lastBatteryLevel = 0
local lastGasolineVolume = 0
local lastBatteryVolume = 0

local function setMode(mode)
  electrics.values.hevState = mode
  if iDM_Transmission.mode ~= mode then
    iDM_Transmission:setMode(mode)
  end
end

local function updateMotorIgnition()
  if electrics.values.ignitionLevel >= 2 then
    for _, motor in ipairs(p1_Motors) do
      motor.ignitionCoef = 1
    end
    for _, motor in ipairs(p3_Motors) do
      motor.ignitionCoef = 1
    end
    for _, motor in ipairs(motors) do
      motor.ignitionCoef = 1
    end
  elseif electrics.values.ignitionLevel < 2 then
    for _, motor in ipairs(p1_Motors) do
      motor.ignitionCoef = 0
    end
    for _, motor in ipairs(p3_Motors) do
      motor.ignitionCoef = 0
    end
    for _, motor in ipairs(motors) do
      motor.ignitionCoef = 0
    end
  end
end

local function updateEHSState(dt)
  electrics.values.checkRE = engine.isDisabled and 1 or 0

  local function startEngine()
    if electrics.values.reRunning == 0 and electrics.values.fuel < electrics.values.hybridSOC / 100 * 0.95 then
      if engine.outputAV1 < rpm1 * rpmToAV * 0.8 then
        engine:deactivateStarter(engine)
        electrics.values.p1RegenThrottle = 0
        electrics.values.p1Throttle = 0.5
      else
        engine:deactivateStarter(engine)
        engine.ignitionCoef = 1
        electrics.values.reRunning = 1
        electrics.values.engineTrottle = throttleSet
        electrics.values.p1Throttle = 0
      end
    end
  end

  local function stopEngine()
    engine:deactivateStarter(engine)
    electrics.values.engineTrottle = 0
    engine.ignitionCoef = 0
    electrics.values.reRunning = 0
    electrics.values.p1Throttle = 0
    electrics.values.p1RegenThrottle = 0
  end

  local function idlingTempKeeper()
    if engine.thermals.coolantTemperature > 90 then
      idlingBreak = 1
    end
    if idlingBreak == 1 and engine.thermals.coolantTemperature < 88 then
      idlingBreak = 0
    end
  end

  if electrics.values.checkRE == 1 then
    electrics.values.hybridMode = "EV"
  end
  if electrics.values.ignitionLevel == 2 and electrics.values.checkRE == 0 and electrics.values.fuel < 0.15 and electrics.values.hybridMode == "EV" then
    electrics.values.hybridMode = "HEV"
  end

  if electrics.values.ignitionLevel < 2 or engine.thermals.coolantTemperature >= 110 or electrics.values.hybridMode == "EV" or electrics.values.hybridMode == "RE" then
    setMode("SERIES")
  elseif electrics.values.ignitionLevel == 2 and electrics.values.hybridMode == "HEV" then
    if electrics.values.fuel >= electrics.values.hybridSOC / 100 or electrics.values.gear ~= "D" then
      setMode("SERIES")
    elseif electrics.values.gear == "D" and electrics.values.fuel < electrics.values.hybridSOC / 100 then
      local p3_OutputAV = 0
      local p3_Count = #p3_Motors
      for _, motor in ipairs(p3_Motors) do
        p3_OutputAV = p3_OutputAV + motor.outputAV1
      end
      if p3_Count > 0 then
        p3_OutputAV = p3_OutputAV / p3_Count
      end
      --if electrics.values.reRunning == 1 and p3_OutputAV >= rpm1 * rpmToAV / iDM_Transmission.gearRatio2 and p3_OutputAV <= freeRPM * rpmToAV / iDM_Transmission.gearRatio2 then
      if electrics.values.reRunning ~= 1 or electrics.values.brake > 0 or p3_OutputAV > freeRPM * rpmToAV / iDM_Transmission.gearRatio2 or (electrics.values.fuel < 0.05 and electrics.values.throttle > throttleSet) then
        setMode("SERIES")
      elseif electrics.values.reRunning == 1 and electrics.values.brake == 0 and p3_OutputAV <= freeRPM * rpmToAV / iDM_Transmission.gearRatio2 then
        if electrics.values.fuel >= 0.05 then
          if p3_OutputAV < rpm1 * rpmToAV / iDM_Transmission.gearRatio2 * 0.95 then
            setMode("SERIES")
          elseif p3_OutputAV > rpm1 * rpmToAV / iDM_Transmission.gearRatio2 * 1.05 then
            setMode("PARALLEL")
          end
        elseif electrics.values.fuel < 0.05 then
          if p3_OutputAV < rpm1 * rpmToAV / iDM_Transmission.gearRatio2 * 0.95 or electrics.values.throttle > throttleSet then
            setMode("SERIES")
          elseif p3_OutputAV > rpm1 * rpmToAV / iDM_Transmission.gearRatio2 * 1.05 and electrics.values.throttle <= throttleSet * 0.9 then
            setMode("PARALLEL")
          end
        end
      end
    end
  end

  if electrics.values.hevState == "PARALLEL" then
    engine:setTempRevLimiter(freeRPM * rpmToAV)
    startEngine()
    electrics.values.p1Throttle = 0
    electrics.values.p1RegenThrottle = 0
    engine.electricsThrottleName = "engineTrottle"
    for _, motor in ipairs(p3_Motors) do
      motor.electricsThrottleName = "parallel_MotorTrottle"
      motor.electricsRegenThrottleName = "parallel_MotorRegenTrottle"
    end
    for _, motor in ipairs(motors) do
      motor.electricsThrottleName = "parallel_MotorTrottle"
      motor.electricsRegenThrottleName = "parallel_MotorRegenTrottle"
    end
    electrics.values.engineTrottle = max(throttleSet, electrics.values.throttle)
    electrics.values.parallel_MotorRegenTrottle = max(0, (throttleSet - electrics.values.throttle) / throttleSet)
    electrics.values.parallel_MotorTrottle = max(0, (electrics.values.throttle - throttleSet - 0.05) / (1 - throttleSet - 0.05))
  else
    for _, motor in ipairs(p3_Motors) do
      motor.electricsThrottleName = "throttle"
      motor.electricsRegenThrottleName = "regenThrottle"
    end
    for _, motor in ipairs(motors) do
      motor.electricsThrottleName = "throttle"
      motor.electricsRegenThrottleName = "regenThrottle"
    end
    engine.electricsThrottleName = "engineTrottle"
    if electrics.values.ignitionLevel < 2 or electrics.values.hybridMode == "EV" or electrics.values.fuel >= electrics.values.hybridSOC / 100 or engine.thermals.coolantTemperature >= 110 then
      idlingTime = 0
      stopEngine()
    elseif electrics.values.ignitionLevel == 2 then
      idlingTempKeeper()
      if electrics.values.wheelspeed * 3.6 <= 15 and electrics.values.fuel > electrics.values.hybridSOC / 100 * 0.5 and idlingBreak == 1 then
        engine:setTempRevLimiter(rpm1 * rpmToAV)
        idlingTime = idlingTime + dt
        if idlingTime >= 15 then
          stopEngine()
        else
          if electrics.values.reRunning == 1 then
            electrics.values.engineTrottle = throttleSet
            if engine.outputAV1 < engine.tempRevLimiterAV * 0.9 then
              electrics.values.p1RegenThrottle = max(0, electrics.values.p1RegenThrottle - 0.02)
            end
            if engine.throttle < throttleSet then
              electrics.values.p1RegenThrottle = min(1, electrics.values.p1RegenThrottle + 0.02)
            end
          else
            electrics.values.engineTrottle = 0
          end
        end
      elseif electrics.values.wheelspeed * 3.6 < speed1 or engine.thermals.coolantTemperature > 100 then
        idlingTime = 0
        if engine.thermals.coolantTemperature > 100 or electrics.values.fuel >= 0.1 then
          engine:setTempRevLimiter(rpm1 * rpmToAV)
        elseif electrics.values.fuel < 0.1 then
          engine:setTempRevLimiter(rpm2 * rpmToAV)
        end
        if engine.thermals.coolantTemperature < 105 then
          startEngine()
        end
        if electrics.values.reRunning == 1 then
          electrics.values.engineTrottle = throttleSet
          if engine.outputAV1 < engine.tempRevLimiterAV * 0.9 then
            electrics.values.p1RegenThrottle = max(0, electrics.values.p1RegenThrottle - 0.02)
          end
          if engine.throttle < throttleSet then
            electrics.values.p1RegenThrottle = min(1, electrics.values.p1RegenThrottle + 0.02)
          end
        else
          electrics.values.engineTrottle = 0
        end
      elseif electrics.values.wheelspeed * 3.6 >= speed2 or electrics.values.fuel <= 0.1 then
        idlingTime = 0
        engine:setTempRevLimiter(rpm3 * rpmToAV)
        startEngine()
        if electrics.values.reRunning == 1 then
          electrics.values.engineTrottle = throttleSet
          if engine.outputAV1 < engine.tempRevLimiterAV * 0.9 then
            electrics.values.p1RegenThrottle = max(0, electrics.values.p1RegenThrottle - 0.02)
          end
          if engine.throttle < throttleSet then
            electrics.values.p1RegenThrottle = min(1, electrics.values.p1RegenThrottle + 0.02)
          end
        else
          electrics.values.engineTrottle = 0
        end
      elseif (electrics.values.wheelspeed * 3.6 >= speed1 and electrics.values.wheelspeed * 3.6 < speed2) or electrics.values.throttle > 0.7 then
        idlingTime = 0
        engine:setTempRevLimiter((rpm2 + (rpm3 - rpm2) * (1 - (speed2 - electrics.values.wheelspeed * 3.6) / (speed2 - speed1))) * rpmToAV)
        startEngine()
        if electrics.values.reRunning == 1 then
          electrics.values.engineTrottle = throttleSet
          if engine.outputAV1 < engine.tempRevLimiterAV * 0.9 then
            electrics.values.p1RegenThrottle = max(0, electrics.values.p1RegenThrottle - 0.02)
          end
          if engine.throttle < throttleSet then
            electrics.values.p1RegenThrottle = min(1, electrics.values.p1RegenThrottle + 0.02)
          end
        else
          electrics.values.engineTrottle = 0
        end
      end
    end
  end
end

local function updateGaugeInfo(dt)
  local totalCapacity = 0
  local remainingEnergy = 0
  local remainingVolume = 0
  
  for _, s in pairs(engine.registeredEnergyStorages) do
    local storage = energyStorage.getStorage(s)
    if storage and storage.type ~= "n2oTank" then
      totalCapacity = totalCapacity + storage.energyCapacity
      remainingEnergy = remainingEnergy + storage.storedEnergy
      remainingVolume = remainingVolume + storage.remainingVolume
    end
  end
  local remainingGasolineRatio = remainingEnergy / totalCapacity
  electrics.values.remainingGasolineRatio = remainingGasolineRatio
  if remainingGasolineRatio <= 0.1 then
    electrics.values.lowGasoline = 1
  else
    electrics.values.lowGasoline = 0
  end

  local motorThrottle = 0
  local motorRegenThrottle = 0
  local motorCount = #p3_Motors + #motors
  electrics.values.motorCount = motorCount
  for _, motor in ipairs(p3_Motors) do
    motorThrottle = motorThrottle + electrics.values[motor.electricsThrottleName]
    motorRegenThrottle = motorRegenThrottle + electrics.values[motor.electricsRegenThrottleName]
  end
  for _, motor in ipairs(motors) do
    motorThrottle = motorThrottle + electrics.values[motor.electricsThrottleName]
    motorRegenThrottle = motorRegenThrottle + electrics.values[motor.electricsRegenThrottleName]
  end
  motorThrottle = motorThrottle / motorCount
  motorRegenThrottle = motorRegenThrottle / motorCount

  if electrics.values.hevState == "PARALLEL" then
    if motorThrottle > 0 then
      electrics.values.hybridEnergyFlow = "P-"
    elseif motorThrottle <= 0 and motorRegenThrottle >= 0 and motorRegenThrottle < 1e-2 then
      electrics.values.hybridEnergyFlow = "PP"
    elseif motorThrottle <= 0 and motorRegenThrottle >= 1e-2 then
      if electrics.values.throttle <= 0 then
        electrics.values.hybridEnergyFlow = "PR"
      else
        electrics.values.hybridEnergyFlow = "P+"
      end
    end
  elseif electrics.values.hevState == "SERIES" then
    if electrics.values.reRunning == 1 and electrics.values.wheelspeed >= 0.1 then
      if motorThrottle > 0 then
        if electrics.values.fuel < lastBatteryLevel then
          electrics.values.hybridEnergyFlow = "S-"
        elseif electrics.values.fuel == lastBatteryLevel then
          electrics.values.hybridEnergyFlow = "SS"
        elseif electrics.values.fuel > lastBatteryLevel then
          electrics.values.hybridEnergyFlow = "S+"
        end
      elseif motorThrottle <= 0 and motorRegenThrottle >= 0 and motorRegenThrottle < 1e-2 then
        electrics.values.hybridEnergyFlow = "SC"
      elseif motorThrottle <= 0 and motorRegenThrottle >= 1e-2 then
        electrics.values.hybridEnergyFlow = "SR"
      end
    elseif electrics.values.reRunning == 0 and electrics.values.wheelspeed >= 0.1 then
      if motorThrottle > 0 then
        electrics.values.hybridEnergyFlow = "E-"
      elseif motorThrottle <= 0 and motorRegenThrottle >= 0 and motorRegenThrottle < 1e-2 then
        electrics.values.hybridEnergyFlow = "EE"
      elseif motorThrottle <= 0 and motorRegenThrottle >= 1e-2 then
        electrics.values.hybridEnergyFlow = "ER"
      end
    elseif electrics.values.wheelspeed < 0.1 then
      if electrics.values.reRunning == 1 then
        electrics.values.hybridEnergyFlow = "SC"
      else
        electrics.values.hybridEnergyFlow = "EE"
      end
    end
  end
  lastBatteryLevel = electrics.values.fuel
end

local function updateGFX(dt)
  updateMotorIgnition()
  updateEHSState(dt)
  updateGaugeInfo(dt)
end

local function serialize()
end

local function deserialize(data)
end

local function init(jbeamData)
  extensions.load("socControl")
  engine = powertrain.getDevice(jbeamData.engineName) or powertrain.getDevice("mainEngine")
  iDM_Transmission = powertrain.getDevice(jbeamData.iDM_Transmission_Name) or powertrain.getDevice(jbeamData.p1Name)
  p1_Motors = powertrain.getDevicesByType("p1_Motor")
  p3_Motors = powertrain.getDevicesByType("p3_Motor")
  motors = powertrain.getDevicesByType("electricMotor")
  freeRPM = engine.maxRPM or 6000
  rpm1 = jbeamData.rpm1 or 1500
  rpm2 = jbeamData.rpm2 or 2000
  rpm3 = jbeamData.rpm3 or 3000
  speed1 = jbeamData.speed1 or 60
  speed2 = jbeamData.speed2 or 100
  lastBatteryLevel = electrics.values.fuel or 0
  electrics.values.checkRE = engine.isDisabled and 1 or 0

  local minV = nil
  for k, v in pairs(engine.invBurnEfficiencyTable) do
    if minV == nil or v < minV or (v == minV and k / 100 < throttleSet) then
      minV = v
      throttleSet = k / 100
    end
  end
end

local function initLastStage(jbeamData)
end

M.init = init
M.initLastStage = initLastStage
M.updateMotorIgnition = updateMotorIgnition
M.updateEHSState = updateEHSState
M.updateGFX = updateGFX
M.serialize = serialize
M.deserialize = deserialize

return M
